/*
	Written by TCH (tch@protonmail.com)
	$VER: 1.2.0
	Last modified: 2020.08.07.
	Public Domain
*/
#ifndef _CONFPARSE_C
#define _CONFPARSE_C

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <dirent.h>

#include "readline.c"

#define CONFVAR_TYPE_STRING	0
#define CONFVAR_TYPE_BOOL	1
#define CONFVAR_TYPE_FLOAT32	2
#define CONFVAR_TYPE_FLOAT64	3
#define CONFVAR_TYPE_INT8	4
#define CONFVAR_TYPE_INT16	5
#define CONFVAR_TYPE_INT32	6
#define CONFVAR_TYPE_INT64	7

#define CONFPARSE_ERROR_FILE_OPEN	1
#define CONFPARSE_ERROR_BAD_NAME	2
#define CONFPARSE_ERROR_BAD_VALUE	3
#define CONFPARSE_ERROR_NO_MEMORY	4
#define CONFPARSE_ERROR_FILE_ERROR	5

#define CONFPARSE_ERROR(cpcode, cpline) *error_code = (cpcode); *error_line = (cpline); fclose(conffile); return;

struct conf_item
{
	char *ci_name;
	int ci_type;
	void *ci_var;
};

char lowercase(char c)
{
	return c + (((c >= 65) && (c <= 90)) << 5);
}

void copytob(char *s, int offset, int stop)
{
	int i;

	i = 0;
	while (offset <= stop)
	{
		s[i++] = s[offset++];
	}
	s[i] = 0;
}

void trimws(char *s)
{
	int i, l, start, stop;

	l = strlen(s);
	i = 0;
	while ((i < l) && (s[i] < 33))
	{
		++i;
	}
	if (i == l)
	{
		s[0] = 0;
		return;
	}
	start = i;

	i = l - 1;
	while ((i > start) && (s[i] < 33))
	{
		--i;
	}
	stop = i;

	copytob(s, start, stop);
}

void parse_conf_file(char *filename, int items_c, struct conf_item items[], int *error_code, int *error_line)
{
	FILE *conffile;
	char c;
	char *line, name[1024], value[1024];
	size_t ll;
	int err, i, j, l, line_no, name_l, value_l, ci_type;
	void *ci_var;
	char *e, *tmp;
	char vc[2];
	bool in_name, value_in_quotes, boolvalue;
	double floatvalue;
	int64_t intvalue;

	*error_code = 0;
	*error_line = -1;

	conffile = fopen(filename, "r");
	if (conffile == NULL)
	{
		*error_code = CONFPARSE_ERROR_FILE_OPEN;
		return;
	}

	vc[false] = 35;
	line_no = 0;
	while (!feof(conffile))
	{
		err = readline(conffile, &line, &ll);
		if (err == -1)
		{
			CONFPARSE_ERROR(CONFPARSE_ERROR_NO_MEMORY, line_no);
		}
		if (err != 0)
		{
			CONFPARSE_ERROR(CONFPARSE_ERROR_FILE_ERROR, line_no);
		}
		l = ll;
		trimws(line);
		if ((strcmp(line, "") != 0) && (line[0] != 35))
		{
			name[0] = 0;
			name_l = 0;
			value[0] = 0;
			value_l = 0;
			in_name = true;
			i = 0;
			l = strlen(line);
			while (i <= l)
			{
				c = lowercase(line[i]);
				if (in_name)
				{
					if (((c >= '0') && (c <= '9')) || ((c >= 'a') && (c <= 'z')) || (c == '_') || (c == '[') || (c == ']') || (c == '.'))
					{
						name[name_l++] = c;
						name[name_l] = 0;
					}
					else
					{
						if (c != '=')
						{
							in_name = false;
							copytob(line, i + 2, l - 1);
							trimws(line);
							vc[true] = line[0];
							value_in_quotes = (line[0] == 34) || (line[0] == 39);
							i = value_in_quotes - 1;
						}
						else
						{
							if ((c != 9) && (c != 32))
							{
								CONFPARSE_ERROR(CONFPARSE_ERROR_BAD_NAME, line_no);
							}
						}
					}
				}
				else
				{
					if (c == vc[value_in_quotes])
					{
						break;
					}
					value[value_l++] = c;
					value[value_l] = 0;
				}
				++i;
			}

			if (strcmp(name, "") != 0)
			{
				i = -1;
				for (j = 0; j < items_c; ++j)
				{
					if (strcmp(items[j].ci_name, name) == 0)
					{
						i = j;
					}
				}
				if (i < 0)
				{
					CONFPARSE_ERROR(CONFPARSE_ERROR_BAD_NAME, line_no);
				}
				ci_type = items[i].ci_type & 7;
				ci_var = items[i].ci_var;
				if (!value_in_quotes)
				{
					trimws(value);
				}

				if (ci_type == CONFVAR_TYPE_STRING)
				{
					tmp = realloc(*((char **)ci_var), strlen(value) + 1);
					if (tmp == NULL)
					{
						CONFPARSE_ERROR(CONFPARSE_ERROR_NO_MEMORY, line_no);
					}
					strcpy(tmp, value);
					*((char **)ci_var) = tmp;
				}
				else
				{
					if ((ci_type == CONFVAR_TYPE_FLOAT32) || (ci_type == CONFVAR_TYPE_FLOAT64))
					{
						errno = 0;
						floatvalue = strtod(value, &e);
						if ((errno != 0) || (value == e))
						{
							CONFPARSE_ERROR(CONFPARSE_ERROR_BAD_VALUE, line_no);
						}
						if (ci_type == CONFVAR_TYPE_FLOAT32)
						{
							*((float *)ci_var) = floatvalue;
						}
						else
						{
							*((double *)ci_var) = floatvalue;
						}
					}
					else
					{
						if ((ci_type >= CONFVAR_TYPE_INT8) && (ci_type <= CONFVAR_TYPE_INT64))
						{
							errno = 0;
							intvalue = strtoll(value, &e, 0);
							if ((errno != 0) || (value == e))
							{
								CONFPARSE_ERROR(CONFPARSE_ERROR_BAD_VALUE, line_no);
							}
							switch (ci_type)
							{
								case CONFVAR_TYPE_INT8:
									*((int8_t *)ci_var) = (int8_t)intvalue;
								break;
								case CONFVAR_TYPE_INT16:
									*((int16_t *)ci_var) = (int16_t)intvalue;
								break;
								case CONFVAR_TYPE_INT32:
									*((int32_t *)ci_var) = (int32_t)intvalue;
								break;
								case CONFVAR_TYPE_INT64:
									*((int64_t *)ci_var) = (int64_t)intvalue;
								break;
							}
						}
						else
						{
							if ((strcmp(value, "true") == 0) || (strcmp(value, "t") == 0) || (strcmp(value, "yes") == 0) || (strcmp(value, "y") == 0))
							{
								boolvalue = true;
							}
							else
							{
								if ((strcmp(value, "false") == 0) || (strcmp(value, "f") == 0) || (strcmp(value, "no") == 0) || (strcmp(value, "n") == 0))
								{
									boolvalue = false;
								}
								else
								{
									errno = 0;
									intvalue = strtoll(value, &e, 0);
									if ((errno != 0) || (value == e))
									{
										CONFPARSE_ERROR(CONFPARSE_ERROR_BAD_VALUE, line_no);
									}
									boolvalue = intvalue;
								}
							}
							*((bool *)ci_var) = boolvalue;
						}
					}
				}
			}
		}
		++line_no;
		free(line);
	}
	fclose(conffile);
}

void scan_conf_dir(char *dirname, int items_c, struct conf_item items[], char **error_file, int *error_code, int *error_line)
{
	char *filename, *tmp;
	size_t dirname_l, entryname_l;
	DIR *d;
	struct dirent *dir;
	struct stat fs;

	*error_file = NULL;
	*error_code = 0;
	*error_line = -1;

	d = opendir(dirname);
	if (d == NULL)
	{
		*error_code = CONFPARSE_ERROR_FILE_OPEN;
		return;
	}

	dirname_l = strlen(dirname);
	dirname_l += (dirname[dirname_l - 1] != '/');
	filename = calloc(dirname_l + 1, 1);
	if (filename == NULL)
	{
		*error_code = CONFPARSE_ERROR_NO_MEMORY;
		closedir(d);
		return;
	}
	strcpy(filename, dirname);
	filename[dirname_l - 1] = '/';
	filename[dirname_l] = 0;

	while ((dir = readdir(d)) != NULL)
	{
		entryname_l = strlen(dir->d_name);
		tmp = realloc(filename, dirname_l + entryname_l + 1);
		if (tmp == NULL)
		{
			*error_code = CONFPARSE_ERROR_NO_MEMORY;
			free(filename);
			closedir(d);
			return;
		}
		filename = tmp;
		strcpy((char *)(filename + dirname_l), dir->d_name);
		filename[dirname_l + entryname_l] = 0;
		stat(filename, &fs);
		if (!S_ISDIR(fs.st_mode))
		{
			parse_conf_file(filename, items_c, items, error_code, error_line);
			if (*error_code != 0)
			{
				if (*error_code == CONFPARSE_ERROR_NO_MEMORY)
				{
					free(filename);
				}
				else
				{
					*error_file = filename;
				}
				closedir(d);
				return;
			}
		}
	}

	free(filename);
	closedir(d);
}


#endif
